unit ImgScaler;

interface

function CalcPitch(Width: Cardinal; Bpp: Cardinal): Cardinal;
procedure Nearest(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Cardinal;
  Bpp: Cardinal; DestWidth, DestHeight: Cardinal;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
procedure Bilinear(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Cardinal;
  Bpp: Cardinal; DestWidth, DestHeight: Cardinal;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
procedure Average(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Cardinal;
  Bpp: Cardinal; DestWidth, DestHeight: Cardinal;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);

implementation

type
  PByte = ^Byte;
  PInteger = ^Integer;
  PTriByte = ^TTriByte;
  TTriByte = array[0..2] of Byte;
  PByteArray = ^TByteArray;
  TByteArray = array[0..MaxInt-1] of Byte;
  PIntArray = ^TIntArray;
  TIntArray = array[0..(MaxInt div SizeOf(Integer))-1] of Integer;
  PPointerArray = ^TPointerArray;
  TPointerArray = array[0..MaxInt div SizeOf(Pointer)-1] of Pointer;

function CalcPitch(Width: Cardinal; Bpp: Cardinal): Cardinal;
begin
  Result := ((Width * Bpp + 7) shr 3 + 3) and not 3;
end;

procedure Nearest(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Cardinal;
  Bpp: Cardinal; DestWidth, DestHeight: Cardinal;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  srcpitch, viewpitch: Integer;
  zx, zy: Double;
  sx: PIntArray;
  sy: PPointerArray;
  sp: PByteArray;
  dline, dp: PByte;
  x, y: Integer;
begin
  srcpitch := CalcPitch(SrcWidth, Bpp);
  viewpitch := CalcPitch(ViewWidth, Bpp);
  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;

  // Set source pixel address
  GetMem(sx, SizeOf(Integer) * ViewWidth);
  for x := 0 to ViewWidth-1 do
    sx[x] := Trunc(zx * (ViewLeft+x+0.5)) * (Bpp div 8);

  GetMem(sy, SizeOf(Pointer) * ViewHeight);
  for y := 0 to ViewHeight-1 do
    sy[y] := @PByteArray(SrcBits)[Trunc(zy * (ViewTop+y+0.5)) * srcpitch];

  dline := DestBits;
  case Bpp of
    8:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          dp^ := sp[sx[x]];
          Inc(dp);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
    24:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          PTriByte(dp)^ := PTriByte(@sp[sx[x]])^;
          Inc(dp, 3);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
  end;

  FreeMem(sx);
  FreeMem(sy);
end;

procedure Bilinear(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Cardinal;
  Bpp: Cardinal; DestWidth, DestHeight: Cardinal;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  fytmp: Integer;
  sx, sy, fx, fy, tmp: PIntArray;
  sp: PByteArray;
  upd: Boolean;
  sp0, sp1, dpt: PTriByte;
  dp: PByte;
  esx, esy, tt: Integer;
  x, y, nextline: Integer;
  dline: PByte;

  srcpitch, viewpitch, c: Integer;
  zx, zy: Double;
  lb: PIntArray;
begin
  srcpitch := CalcPitch(SrcWidth, Bpp);
  viewpitch := CalcPitch(ViewWidth, Bpp);
  c := Bpp div 8;
  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;

  // Set source pixel address
  GetMem(sx, SizeOf(Integer) * ViewWidth);
  GetMem(fx, SizeOf(Integer) * ViewWidth);
  for x := 0 to ViewWidth-1 do
  begin
    tt := Trunc((zx * (ViewLeft+x+0.5) - 0.5) * $100);
    if tt < 0 then
      tt := 0;
    sx[x] := (tt shr 8) * c;
    fx[x] := tt and $FF;
  end;

  GetMem(sy, SizeOf(Integer) * ViewHeight);
  GetMem(fy, SizeOf(Integer) * ViewHeight);
  for y := 0 to ViewHeight-1 do
  begin
    tt := Trunc((zy * (ViewTop+y+0.5) - 0.5) * $100);
    if tt < 0 then
      tt := 0;
    sy[y] := (tt shr 8) * srcpitch;
    fy[y]  := tt and $FF;
  end;

  GetMem(lb, SizeOf(Integer) * viewpitch * 2);

  esx := (Integer(SrcWidth)-1) * c;
  esy := (Integer(SrcHeight)-1) * srcpitch;
  dline := DestBits;
  //  +---+---+
  //  |sp2|sp3|  fy
  //  +---+---+  ^
  //  |sp0|sp1|  |
  //  +---+---+  +->fx
  // sp0*(1-fx)*(1-fy) + sp1*fx*(1-fy) + sp2*(1-fx)*fy + sp3*fx*fy;
  // sp2: sp0+nextline, sp3: sp1+nextline
  case Bpp of
    8:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := PByteArray(@PByteArray(SrcBits)[sy[y]]);
        if sy[y] < esy then
          nextline := srcpitch
        else
          nextline := 0;
        tmp := lb;
        upd := (y = 0) or (sy[y] <> sy[y-1]);
        fytmp := fy[y];
        x := 0;
        while x < ViewWidth do // "while" faster than "for"...
        begin
          if upd then
          begin
            sp0 := PTriByte(@sp[sx[x]]);
            if sx[x] < esx then
              sp1 := PTriByte(@PByteArray(sp0)[1])
            else
              sp1 := sp0;
            tmp[0] := (sp0[0] shl 8 - (sp0[0] - sp1[0]) * fx[x]);
            sp0 := PTriByte(@sp0[nextline]);
            sp1 := PTriByte(@sp1[nextline]);
            tmp[1] := (sp0[0] shl 8 - (sp0[0] - sp1[0]) * fx[x]);
          end;

          dp^ := (tmp[0] shl 8 - (tmp[0] - tmp[1]) * fytmp) shr 16;
          Inc(PInteger(tmp), 2);

          Inc(dp);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
    24:
      for y := 0 to ViewHeight-1 do
      begin
        dpt := PTriByte(dline);
        sp := PByteArray(@PByteArray(SrcBits)[sy[y]]);
        if sy[y] < esy then
          nextline := srcpitch
        else
          nextline := 0;
        tmp := lb;
        upd := (y = 0) or (sy[y] <> sy[y-1]);
        fytmp := fy[y];
        x := 0;
        while x < ViewWidth do
        begin
          if upd then
          begin
            sp0 := PTriByte(@sp[sx[x]]);
            if sx[x] < esx then
              sp1 := PTriByte(@PByteArray(sp0)[3])
            else
              sp1 := sp0;
            // transform: tmp[0] := (sp0[0] * ($100-fx[x]) + sp1[0] * fx[x]);
            tmp[0] := (sp0[0] shl 8 - (sp0[0] - sp1[0]) * fx[x]);
            tmp[1] := (sp0[1] shl 8 - (sp0[1] - sp1[1]) * fx[x]);
            tmp[2] := (sp0[2] shl 8 - (sp0[2] - sp1[2]) * fx[x]);
            sp0 := PTriByte(@sp0[nextline]);
            sp1 := PTriByte(@sp1[nextline]);
            tmp[3] := (sp0[0] shl 8 - (sp0[0] - sp1[0]) * fx[x]);
            tmp[4] := (sp0[1] shl 8 - (sp0[1] - sp1[1]) * fx[x]);
            tmp[5] := (sp0[2] shl 8 - (sp0[2] - sp1[2]) * fx[x]);
          end;

          dpt[0] := (tmp[0] shl 8 - (tmp[0] - tmp[3]) * fytmp) shr 16;
          dpt[1] := (tmp[1] shl 8 - (tmp[1] - tmp[4]) * fytmp) shr 16;
          dpt[2] := (tmp[2] shl 8 - (tmp[2] - tmp[5]) * fytmp) shr 16;
          Inc(PInteger(tmp), 6);

          Inc(dpt);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
  end;

  FreeMem(lb);
  FreeMem(sx);
  FreeMem(sy);
  FreeMem(fx);
  FreeMem(fy);
end;

procedure Average(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Cardinal;
  Bpp: Cardinal; DestWidth, DestHeight: Cardinal;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  d, f: array[0..2] of Cardinal;
  db, dg, dr, fb ,fg, fr, nx ,ny: Cardinal;
  sp: PByteArray;
  spt, dpt: PTriByte;
  i, j, yy, ssx, ssy: Integer;
  yb, ybuf: PIntArray;
  sline, sp0, dline, dp: PByte;
  px, py, px2, py2: Word;
  sx, sy, fx, fy: PIntArray;
  x, y: Integer;

  lastline: Boolean;
  srcPitch, viewpitch: Integer;
  zx, zy: Double;
begin
  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;

  // Set source pixel coordinates
  // Since the difference of the current and following pixel is
  // required, table size is width+1.
  GetMem(sx, SizeOf(Integer) * ViewWidth + 1);
  GetMem(fx, SizeOf(Integer) * ViewWidth + 1);
  for x := 0 to ViewWidth do
  begin
    sx[x] := Trunc((zx*(ViewLeft+x)) * $10000) shr 16;
    fx[x] := Trunc((zx*(ViewLeft+x)) * $10000) and $FFFF;
  end;
  GetMem(sy, SizeOf(Integer) * ViewHeight + 1);
  GetMem(fy, SizeOf(Integer) * ViewHeight + 1);
  for y := 0 to ViewHeight do
  begin
    sy[y] := Trunc((zy*(ViewTop+y)) * $10000) shr 16;
    fy[y] := Trunc((zy*(ViewTop+y)) * $10000) and $FFFF;
  end;

  nx := Trunc($10000 / zx);
  ny := Trunc($10000 / zy);

  // ybuf: buffer of pixels averaged horizontally
  GetMem(ybuf, (Trunc(zy)+2)*SizeOf(Integer)*(Bpp div 8));

  srcpitch := CalcPitch(SrcWidth, Bpp);
  viewpitch := CalcPitch(ViewWidth, Bpp);
  dline := DestBits;

  case Bpp of
    8:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := PByteArray(@PByteArray(SrcBits)[sy[y]*srcpitch]);
        yy := sy[y+1] - sy[y];
        py := $10000 - fy[y];
        py2 := fy[y+1];
        ssy := yy-Ord(py<>0);
        lastline := ViewTop+y = Integer(DestHeight)-1;

        x := 0;
        while x < ViewWidth do
        begin
          yb := ybuf;
          px := $10000 - fx[x];
          px2 := fx[x+1];
          ssx := sx[x+1]-sx[x]-Ord(px<>0);
          sline := @sp[sx[x]];
          for i := 0 to yy do
          begin
            sp0 := sline;

            d[0] := 0;
            f[0] := 0;

            if px <> 0 then
            begin
              Inc(f[0], sp0^ * px);
              Inc(sp0);
            end;
            j := 0;
            while j < ssx do
            begin
              Inc(d[0], sp0^);
              Inc(sp0);
              Inc(j);
            end;
            if px2 <> 0 then
            begin
              Inc(f[0], sp0^ * px2);
            end;

            // When view-size is small, error is large if truncated,
            // so add 0x8000 in order to round off.
            yb[0] := ((d[0] + f[0] shr 16) * nx + $8000) shr 16;
            yb := PIntArray(@yb[1]);

            if not lastline then
              Inc(sline, srcpitch);
          end;

          db:= 0;
          fb := 0;

          yb := ybuf;
          if py <> 0 then
          begin
            Inc(fb, yb[0] * py);
            yb := PIntArray(@yb[1]);
          end;
          j := 0;
          while j < ssy do
          begin
            Inc(db, yb[0]);
            yb := PIntArray(@yb[1]);
            Inc(j);
          end;
          if py2 <> 0 then
          begin
            Inc(fb, yb[0] * py2);
          end;

          // When view-size is small, error is large if truncated,
          // so add 0x8000 in order to round off.
          dp^ := ((db + fb shr 16) * ny + $8000) shr 16;
          Inc(dp);

          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
    24:
      for y := 0 to ViewHeight-1 do
      begin
        dpt := PTriByte(dline);
        sp := PByteArray(@PByteArray(SrcBits)[sy[y]*srcpitch]);
        yy := sy[y+1] - sy[y];
        py := $10000 - fy[y];
        py2 := fy[y+1];
        ssy := yy - Ord(py<>0);
        lastline := ViewTop+y = Integer(DestHeight)-1;

        x := 0;
        while x < ViewWidth do
        begin
          yb := ybuf;
          px := $10000 - fx[x];
          px2 := fx[x+1];
          ssx := sx[x+1] - sx[x] - Ord(px<>0);
          sline := @sp[sx[x]*3];
          for i := 0 to yy do
          begin
            spt := PTriByte(sline);

            d[0] := 0;
            d[1] := 0;
            d[2] := 0;
            f[0] := 0;
            f[1] := 0;
            f[2] := 0;

            if px <> 0 then
            begin
              Inc(f[0], spt[0] * px);
              Inc(f[1], spt[1] * px);
              Inc(f[2], spt[2] * px);
              Inc(spt);
            end;
            j := ssx;
            while j > 0 do
            begin
              Inc(d[0], spt[0]);
              Inc(d[1], spt[1]);
              Inc(d[2], spt[2]);
              Inc(spt);
              Dec(j);
            end;
            if px2 <> 0 then
            begin
              Inc(f[0], spt[0] * px2);
              Inc(f[1], spt[1] * px2);
              Inc(f[2], spt[2] * px2);
            end;

            // When view-size is small, error is large if truncated,
            // so add 0x8000 in order to round off.
            yb[0] := ((d[0] + f[0] shr 16) * nx + $8000) shr 16;
            yb[1] := ((d[1] + f[1] shr 16) * nx + $8000) shr 16;
            yb[2] := ((d[2] + f[2] shr 16) * nx + $8000) shr 16;
            yb := PIntArray(@yb[3]);

            if not lastline then
              Inc(sline, srcpitch);
          end;

          db := 0;
          dg := 0;
          dr := 0;
          fb := 0;
          fg := 0;
          fr := 0;
          yb := ybuf;
          if py <> 0 then
          begin
            Inc(fb, yb[0] * py);
            Inc(fg, yb[1] * py);
            Inc(fr, yb[2] * py);
            yb := PIntArray(@yb[3]);
          end;
          j := 0;
          while j < ssy do
          begin
            Inc(db, yb[0]);
            Inc(dg, yb[1]);
            Inc(dr, yb[2]);
            yb := PIntArray(@yb[3]);
            Inc(j);
          end;
          if py2 <> 0 then
          begin
            Inc(fb, yb[0] * py2);
            Inc(fg, yb[1] * py2);
            Inc(fr, yb[2] * py2);
          end;

          // When view-size is small, error is large if truncated,
          // so add 0x8000 in order to round off.
          dpt[0] := ((db + fb shr 16) * ny + $8000) shr 16;
          dpt[1] := ((dg + fg shr 16) * ny + $8000) shr 16;
          dpt[2] := ((dr + fr shr 16) * ny + $8000) shr 16;
          Inc(dpt);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
  end;

  FreeMem(sx);
  FreeMem(sy);
  FreeMem(fx);
  FreeMem(fy);
  FreeMem(ybuf);
end;

end.
