unit NCommand;

interface

uses Windows, NDict;

type
  TCommandMethod = procedure of object;

  TCommand = class(TObject)
    AccelHandle: HACCEL;
    Commands: array of Pointer;
    Obj: TObject;
    constructor Create(const Obj: TObject;
        const Commands: array of TCommandMethod);
    destructor Destroy; override;
    function Id(const Name: string): Integer; overload;
    function Id(const Method: TCommandMethod): Integer; overload;
    procedure AddMenuItems(MenuHandle: HMENU; const Items: array of string);
    procedure BindKeys(const KeyMap: TStrArrayDict);
    procedure DispatchCommand(Id: Integer);
  end;

implementation

uses NKeyMap;

const
  FirstId = 20000;
  CommandMethodPrefix = 'cm';

constructor TCommand.Create(const Obj: TObject;
    const Commands: array of TCommandMethod);
var
  i: Integer;
begin
  Self.Obj := Obj;
  SetLength(Self.Commands, Length(Commands));
  for i := 0 to High(Self.Commands) do
    Self.Commands[i] := TMethod(Commands[i]).Code;
end;

destructor TCommand.Destroy;
begin
  if AccelHandle <> 0 then
    DestroyAcceleratorTable(AccelHandle);
end;

procedure TCommand.AddMenuItems(MenuHandle: HMENU;
    const Items: array of string);
  function AccelToKeyName(const Accel: TAccel): string;
  var
    ks: TKeyState;
  begin
    ks := [];
    if (Accel.fVirt and FALT)     <> 0 then Include(ks, ksAlt);
    if (Accel.fVirt and FSHIFT)   <> 0 then Include(ks, ksShift);
    if (Accel.fVirt and FCONTROL) <> 0 then Include(ks, ksCtrl);
    Result := VkeyToKeyName(ks, Accel.key);
  end;
var
  i, j, id: Integer;
  capt, name, keyname: string;
  accels: array of TAccel;
  menu: HMENU;
begin
  menu := MenuHandle;
  SetLength(accels, CopyAcceleratorTable(AccelHandle, Pointer(nil)^, 0));
  CopyAcceleratorTable(AccelHandle, accels[0], Length(accels));

  i := 0;
  while i <= High(Items) do
  begin
    capt := Items[i];
    name := Items[i+1];
    if name = '' then
      menu := MenuHandle
    else if name = '-' then
      AppendMenu(menu, MF_SEPARATOR, 0, nil)
    else if name[1] = '>' then
    begin
      menu := CreatePopupMenu;
      AppendMenu(MenuHandle, MF_POPUP, menu, PChar(capt));
    end else
    begin
      keyname := '';
      id := Self.Id(name);
      for j := 0 to High(accels) do
        if accels[j].cmd = id then
          if keyname = '' then
            keyname := AccelToKeyName(accels[j])
          else
            keyname := keyname + ', ' + AccelToKeyName(accels[j]);
      AppendMenu(menu, MF_STRING, id, PChar(capt + #9 + keyname));
    end;
    Inc(i, 2);
  end;
end;

procedure TCommand.Bindkeys(const KeyMap: TStrArrayDict);
var
  accels: array of TAccel;
  i: Integer;
  key: string;
  ks: TKeyState;
  virt: Byte;
  vkey: Word;
begin
  SetLength(accels, KeyMap.Count);
  i := 0;
  while KeyMap.EachKeys(key) do
    if Length(KeyMap[key]) > 0 then
    begin
      StrToVkey(key, ks, vkey);
      virt := FVIRTKEY;
      if ksAlt   in ks then virt := virt or FALT;
      if ksCtrl  in ks then virt := virt or FCONTROL;
      if ksShift in ks then virt := virt or FSHIFT;
      accels[i].key := vkey;
      accels[i].fVirt := virt;
      accels[i].cmd := Id(KeyMap[key][0]);
      Inc(i);
    end;
  if AccelHandle <> 0 then
    DestroyAcceleratorTable(AccelHandle);
  AccelHandle := CreateAcceleratorTable(accels[0], Length(accels));
end;

function TCommand.Id(const Name: string): Integer;
var
  i: Integer;
begin
  for i := 0 to High(Commands) do
    if Commands[i] = Obj.MethodAddress(CommandMethodPrefix+Name) then
    begin
      Result := FirstId + i;
      Exit;
    end;
  Result := 0;
end;

function TCommand.Id(const Method: TCommandMethod): Integer;
var
  i: Integer;
begin
  for i := 0 to High(Commands) do
    if Commands[i] = TMethod(Method).Code then
    begin
      Result := FirstId + i;
      Exit;
    end;
  Result := 0;
end;

procedure TCommand.DispatchCommand(Id: Integer);
var
  m: TMethod;
begin
  m.Data := Obj;
  m.Code := Commands[Id - FirstId];
  TCommandMethod(m);
end;

end.
